/*************************************************************************\
 * File   : TCLTools.cpp
 * Purpose: Class wrapper for some internal TCL commands
 *
 * Author : Tom Wilkason
 * Dated  : 8/22/1999
\************************************************************************/
#ifndef TCL_LIST_CLASS
#define TCL_LIST_CLASS
/**
 * Enable the TCL_DECLARE_MUTEX for semaphore
 *
 */
#include <tcl.h>
//////////////////////////////////////////////////////////////////////
// TCL Dynamic Object Fast Class
//////////////////////////////////////////////////////////////////////
/*
 This class does not call Tcl_ListObjAppendElement for each new item
 to prevent the continual checks and reallocs that occur using this
 method to expand lists dynamically. The idea here is to:
 - At sublist start, each level will initialize itself once at INIT_LIST_DEPTH elements.
   This also applies to level 0 (the base level).
 - At a sublist end, a new list object will be created which copies
   each of the object array pointer items to it. The internal list
   can then be reused.
 - When adding a new object, check the current size of the internal
   array and if not large enough, realloc at double the size. The idea
   is that this won't have to happen very often.
 - During destruction, all the arrays will be deallocated
 - When setting the result, create a new list object at the base level,
   then use it to set the result.
*/
#define INIT_LIST_DEPTH 100
#define MAX_TDO_DEPTH 100
#define TDOcl_DUMMY (struct Tcl_Obj *)12321
/*
 * Structure to hold the list array stuff, one of these per
 * nested level.
 */
typedef struct stTIO {
   int depth;
   int index;
   Tcl_Obj **obList;
} stTIO;
/*
 * Tcl Dynamic Object Fast Class (TDOfcl)
 */
class TDOfcl
{
public:
   /*
    * Constructor (def to no iterp)
    */
   TDOfcl(Tcl_Interp *Interp=0)
   {
      tdo[0] = TDOcl_DUMMY;
      max_tdo_sofar = 0;
      dmpsInterp = 0;
      /*
       * Allocate a list of initial object pointers
       */
      tio[0].obList = (Tcl_Obj **)Tcl_Alloc(INIT_LIST_DEPTH* sizeof(Tcl_Obj *));
      if (tio[0].obList == NULL) {
         panic("TDOfcl::DSInit: Could not allocate memory!");
      }
      tio[0].depth = INIT_LIST_DEPTH;
      tio[0].index = 0;
      // Clear out remaining list data
      for (int i=1; i < MAX_TDO_DEPTH; i++) {
         tio[i].index = 0;
         tio[i].depth = 0;
         tio[i].obList = NULL;
      }
      asyncMutex = NULL;
      Init(Interp);
   };
   /*
    * Destructor, remove ref count on base list object (if not freed)
    * and release all of the memory for each possible nested level
    */
   ~TDOfcl()
   {
      if (tdo[0]!=TDOcl_DUMMY)
      {
         Tcl_DecrRefCount(tdo[0]);
         tdo[0] = TDOcl_DUMMY;
      }
      /*
       * Remove all of the alloc'd memory  MAX_TDO_DEPTH
       */
      for (int i=0; i <= max_tdo_sofar; i++) {
         if (tio[i].obList != NULL)
         {
            Tcl_Free((char *)tio[i].obList);
            tio[i].depth = 0;
            tio[i].index = 0;
         }
      }
   };
   /*
    * Reset the info and create a new base
    * list object if one is needed (say, after a set result).
    * This is normally not needed externally.
    */
   void Init(Tcl_Interp *interp=0)
   {
      if (interp > 0)
      {
         dmpsInterp=interp;
      }
      itdo=0;
      Elements = 0;
      Semaphore = 0;
      tries = 0;
      /**
       * If a base handle doesn't exist, then create one
       */
      if ( tdo[0] == TDOcl_DUMMY ) {
         tdo[0]= Tcl_NewListObj(0, NULL);
      }
   };
   /**************************************************************
    *               Methods to add elements                      *
    **************************************************************/
   /*
    * Add a string
    */
   void TDOfcl::Add(char *string) {
      addItem(Tcl_NewStringObj(string,-1));
   };
   /*
    * Append an int
    */
   void TDOfcl::Add(int value)
   {
      addItem(Tcl_NewIntObj (value));
   };
   /*
    * Append a long
    */
   void TDOfcl::Add(long value)
   {
      addItem(Tcl_NewIntObj (value));
   };
   /*
    * Append a double
    */
   void TDOfcl::Add(double value)
   {
      addItem(Tcl_NewDoubleObj (value));
   };
   /*
    * Append a float
    */
   void TDOfcl::Add(float value)
   {
      addItem(Tcl_NewDoubleObj (value));
   };
   /*
    * Append an unsigned int
    */
   void TDOfcl::Add(unsigned int value)
   {
      addItem(Tcl_NewLongObj (value));
   };

   /*****************************************************************
    * Start a nested list level by allocating a buffer if necessary *
    *****************************************************************/
   void TDOfcl::SubStart()
   {
      /**
       * Create a new list object one up the stack
       */
      if (itdo < MAX_TDO_DEPTH-1)
      {
         itdo++;
         // Track max depth used for cleanup
         if (itdo > max_tdo_sofar) {
            max_tdo_sofar = itdo;
         }

         //int index = tio[itdo].index;
         /*
          * If necessary, initialize the object array at INIT_LIST_DEPTH elementes
          */
         if (tio[itdo].depth == 0) {
            tio[itdo].obList = (Tcl_Obj **)Tcl_Alloc(INIT_LIST_DEPTH* sizeof(Tcl_Obj *));
            if (tio[itdo].obList == NULL) {
               panic("TDOfcl::DSSubStart: Could not allocate memory!");
            }
            tio[itdo].depth =INIT_LIST_DEPTH;
         }
      }
   };
   /*
    * Complete the nest level by creating a new list from the curren buffer and appending
    * it to the level beneath.
    */
   void TDOfcl::SubEnd()
   {
      /*
       * Create a new list object copying list array over
       * Append the object pointer from one level above to one below
       */
      if (itdo > 0)
      {
         tdo[itdo]= Tcl_NewListObj(tio[itdo].index, tio[itdo].obList);         // (Start new sublist) (elist is pointer to it)
         tio[itdo].index = 0;
         itdo--;     // Work on one up from here
         addItem(tdo[itdo+1]);
      }
   };
   /*
    * Set the result pointer to the nested list structure and reset for the next go
    */
   void TDOfcl::SetResult()
   {
      if (dmpsInterp != NULL) {
         if (tdo[0]!=TDOcl_DUMMY)
         {
            Tcl_DecrRefCount(tdo[0]);
         }
         tdo[0]= Tcl_NewListObj(tio[0].index, tio[0].obList);         // (Start new sublist) (elist is pointer to it)
         Tcl_SetObjResult(dmpsInterp,tdo[0]);          // This frees the handle on top level stack
      }
      tdo[0] = TDOcl_DUMMY;                                // We don't track the handle any more
      tio[0].index = 0;

      // Create a new handle one for next time (Static case)
      Init(dmpsInterp);
   };

   /*****************************************************************
    *                        Mutex Stuff                            *
    *****************************************************************/
   /*
    * Grab a lock with a user ID
    */
   int TDOfcl::Grab(int User=0)
   {
      Tcl_MutexLock(&asyncMutex);
      Semaphore = User;
      Elements++;
      return 0;

   };
   /*
    * Release the lock
    */
   void TDOfcl::Release()
   {
      Semaphore = 0;
      Tcl_MutexUnlock(&asyncMutex);
   };

   int       Semaphore;
   int       ThisDepth;
   int       tries;
   int       Elements;
private:
   /*
    * Add the new object to the internal list structure
    */
   void TDOfcl::addItem(Tcl_Obj *objPtr)
   {
      /*
       * If current level has space, add item, otherwise increase space and add
       */
      /*
       * Double the memory if needed
       */
      int index = tio[itdo].index;
      if ( (index + 1) >= tio[itdo].depth ) {
         tio[itdo].obList = (Tcl_Obj **)Tcl_Realloc((char *)tio[itdo].obList,
            (tio[itdo].depth*2)* sizeof(Tcl_Obj *)
         );
         if (tio[itdo].obList == NULL) {
            panic("TDOfcl::addItem: Could not allocate memory!");
         }

         tio[itdo].depth *=2;
      }
      tio[itdo].obList[index] = objPtr;
      tio[itdo].index++;     // Point to the next free entry
   };

   Tcl_Mutex asyncMutex;
   int       HasInterp;
   int       IsReady;
   int       max_tdo_sofar;
   Tcl_Obj  *tdo[MAX_TDO_DEPTH];  /* List pointers */
   stTIO     tio[MAX_TDO_DEPTH];  /* Internal array pointers */
   int       itdo;                /* current nest level pointer */
   Tcl_Interp         *dmpsInterp;
};

///////////////////////////////////////////////////////////////////////
//        TCL Linked Variable Class
///////////////////////////////////////////////////////////////////////
class TLVcl
{
public:
   TLVcl(Tcl_Interp *Interp, int method);
   TLVcl();
   TLVInit(Tcl_Interp *Interp, int method);
   int GetInt(char *variable);
   int GetInt(char *variable, char *index);
   int GetHex(char *variable);
   int GetHex(char *variable, char *index);
   char *GetString(char *variable, char *index);
   int Set(char *variable, char *global, char *index);
   int Set(float variable, char *global, char *index);
   int Set(double variable, char *global, char *index);
   int Set(unsigned short variable, char *global, char *index);
   int Set(int variable, char *global, char *index);
   float GetFloat(char *variable, char *index);
   double GetDouble(char *variable, char *index);
   // mutex stuff
   Tcl_Mutex  tlvAsyncMutex;
   int  Semaphore;
   int TLVcl::Grab(int User)
   {
      Tcl_MutexLock(&tlvAsyncMutex);
      Semaphore = User;
      return 0;

   };
   void TLVcl::Release()
   {
      Semaphore = 0;
      Tcl_MutexUnlock(&tlvAsyncMutex);
   };

private:
   Tcl_Interp       *MyInterp;
   int               MyMethod;
   int               isReady;
   char Value[60];

};
/** TFW **
 * Mutex class
 *
 */
class Mutex
{
public:
   void Mutex::Grab(void)
   {
      //Tcl_MutexLock(&muAsyncMutex);
   };
   void Mutex::Release(void)
   {
      //Tcl_MutexUnlock(&muAsyncMutex);
   };
private:
   Tcl_Mutex muAsyncMutex;
   //Tcl_Mutex  muAsyncMutex;
};

///////////////////////////////////////////////////////////////////////
//        TCL Linked Variable Routines
///////////////////////////////////////////////////////////////////////
int hexTcl_GetVar ( Tcl_Interp *interp, char *Var, int flags );
int intTcl_GetVar ( Tcl_Interp *interp, char *Var, int flags );
int intTcl_SetVar ( Tcl_Interp *interp, char *Var, int intValue, int flags );
int numTcl_SetVar ( Tcl_Interp *interp, char *Var, int Pos, char *Value, int flags );
int decTcl_SetVar ( Tcl_Interp *interp, char *Var, int Pos, int intValue, int flags );

#endif



